/* Client process that communicates with GNU Emacs acting as server.
   Copyright (C) 1986, 1987 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is distributed in the hope that it will be useful,
but without any warranty.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.

Everyone is granted permission to copy, modify and redistribute
GNU Emacs, but only under the conditions described in the
document "GNU Emacs copying permission notice".   An exact copy
of the document is supposed to have been given to you along with
GNU Emacs so that you can know how you may redistribute it all.
It should be in a file named COPYING.  Among other things, the
copyright notice and this notice must be preserved on all copies.  */


#define NO_SHORTNAMES
#include "../src/config.h"
#undef read
#undef write
#undef open
#ifdef close
#undef close
#endif


#if !defined(BSD) && !defined(HAVE_SYSVIPC) && !defined(QNX)
#include <stdio.h>

main ()
{
  fprintf (stderr, "Sorry, the Emacs server is supported only on Berkeley Unix\n");
  fprintf (stderr, "or System V systems with IPC\n");
  exit (1);
}

#else /* BSD or HAVE_SYSVIPC or QNX */

#if defined(BSD) && ! defined (HAVE_SYSVIPC)
/* BSD code is very different from SYSV IPC code */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>

main (argc, argv)
     int argc;
     char **argv;
{
  int s, n, i;
  FILE *out;
  struct sockaddr_un server;
  char *homedir, *cwd, *str;
  char string[BUFSIZ];

  char *getenv (), *getwd ();

  if (argc < 2)
    {
      printf ("Usage: %s [filename]\n", argv[0]);
      exit (1);
    }

  /* 
   * Open up an AF_UNIX socket in this person's home directory
   */

  if ((s = socket (AF_UNIX, SOCK_STREAM, 0)) < 0)
    {
      perror ("socket");
      exit (1);
    }
  server.sun_family = AF_UNIX;
  if ((homedir = getenv ("HOME")) == NULL)
    {
      fprintf (stderr, "No home directory\n");
      exit (1);
    }
  strcpy (server.sun_path, homedir);
  strcat (server.sun_path, "/.emacs_server");
  if (connect (s, &server, strlen (server.sun_path) + 2) < 0)
    {
      perror ("connect");
      exit (1);
    }
  if ((out = fdopen (s, "r+")) == NULL)
    {
      perror ("fdopen");
      exit (1);
    }

  cwd = getwd (string);
  if (cwd == 0)
    abort ();

  for (i = 1; i < argc; i++)
    {
      if (*argv[i] == '+')
	{
	  char *p = argv[i] + 1;
	  while (*p >= '0' && *p <= '9') p++;
	  if (*p != 0)
	    fprintf (out, "%s/", cwd);
	}
      else if (*argv[i] != '/')
	fprintf (out, "%s/", cwd);
      fprintf (out, "%s ", argv[i]);
    }
  fprintf (out, "\n");
  fflush (out);

  printf ("Waiting for Emacs...");
  fflush (stdout);

  rewind (out); /* re-read the output */
  /* str = fgets (string, BUFSIZ, out); */

  /* Now, wait for an answer and print any messages.  */
  
  while (str = fgets (string, BUFSIZ, out))
    printf ("%s", str);
  
  exit (0);
}

#else /* This is the SYSV IPC section */
#ifdef QNX

#include <stdio.h>
#include <stdlib.h>
#include <sys/kernel.h>
#include <unistd.h>
#include <string.h>
#include "/usr/include/pwd.h"

static char pathbuf[PATH_MAX];

static char xmtbuf[4096];
static char rcvbuf[1024];
static char *progname;

main(int argc, char *argv[])
{
  int lastargc, len, lastlen, arglen, left;
  char **lastargv;
  char *ptr, *lastptr, *argptr;
  pid_t serverid;
  struct passwd *pw;
  progname=*argv++;
  argc--;
  if(argc <= 0) {
    fprintf(stderr, "Usage: %s [+<num> | file]...\n", progname);
    exit(1);
  }
/*
Construct the server name as /gnu/emacs/server/<user>
As qnx_name_locate searches the local node first, and as most users
run only one copy of emacs (or, at least, one copy of the server), this
should be enough to identify the server.
*/
  pw=getpwuid(getuid());
  if(!pw) {
    perror("getpwuid");
    exit(1);
  }
  if(!pw->pw_name || !pw->pw_name[0]) {
    fputs("What, no name?  Fix the passwd file!\n", stderr);
    exit(1);
  }
  sprintf(pathbuf, "/gnu/emacs/server/%s", pw->pw_name);
  serverid = qnx_name_locate(0, pathbuf, 0, NULL);
  if (serverid < 0) {
    fprintf(stderr, "%s: qnx_name_locate: ", progname);
    perror(pathbuf);
    exit(1);
  }
  lastargc=argc;
  lastargv=argv;
  ptr = lastptr = xmtbuf;
  len = lastlen = 0;
  left = sizeof(xmtbuf);
  while(argc) {
    while(argc) {
      if (**argv == '+')
	argptr = *argv;
      else if(qnx_fullpath(pathbuf, *argv))
	argptr = pathbuf;
      else {
	fprintf(stderr, "%s: ", progname);
	perror(*argv);
	argv++;
	argc--;
	continue;
      }
      if ((arglen=strlen(argptr)) + 3 < left) {	/* Leave space for " \n\0" */
	memcpy(ptr, argptr, arglen);
	ptr += arglen;
	len += arglen + 1;
	left -= arglen + 1;
	*ptr++=' ';
	/* make sure that +num file occurs in a single chunk */
	if(**argv != '+') {
	  lastargc=argc;
	  lastargv=argv;
	  lastptr = ptr;
	  lastlen = len;
	}
	argv++;
	argc--;
      } else {
	/* We're out of buffer space, revert to after the last file */
	argc=lastargc;
	argv=lastargv;
	ptr = lastptr;
	len = lastlen;
	break;
      }
    }
    *ptr++ = '\n';		/* Add newline */
    *ptr = '\0';		/* Add null */
    len += 2;			/* Include both */
    printf("Waiting for Emacs...");
    fflush (stdout);
    /* Send the list of files */
    if(Send(serverid, xmtbuf, rcvbuf, len, sizeof(rcvbuf)) < 0) {
      fprintf(stderr, "%s: ", progname);
      perror("Send");
      exit(1);
    }
    /* Retreive the reply and print it (usually "Done\n") */
    fputs(rcvbuf, stdout);
    ptr = xmtbuf;
    len = 0;
    left = sizeof(xmtbuf);
  }
  exit(0);
}
#else

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <stdio.h>

main (argc, argv)
     int argc;
     char **argv;
{
  int s;
  key_t key;
  struct msgbuf * msgp =
      (struct msgbuf *) malloc (sizeof *msgp + BUFSIZ);
  struct msqid_ds * msg_st;
  char *homedir, *getenv (), buf[BUFSIZ];
  char *getwd (), *getcwd (), gwdirb[BUFSIZ], *cwd;
  if (argc < 2)
    {
      printf ("Usage: %s [filename]\n", argv[0]);
      exit (1);
    }

  /*
   * Create a message queue using ~/.emacs_server as the path for ftok
   */
  if ((homedir = getenv ("HOME")) == NULL)
    {
      fprintf (stderr,"No home directory\n");
      exit (1);
    }
  strcpy (buf, homedir);
  strcat (buf, "/.emacs_server");
  creat (buf, 0600);
  key = ftok (buf, 1);	/* unlikely to be anyone else using it */
  s = msgget (key, 0600);
  if (s == -1)
    {
      perror ("msgget");
      exit (1);
    }

  msgp->mtext[0] = 0;
  argc--; argv++;
  while (argc)
    {
      if (*argv[0] != '/')
	{
	  char *val;
	  cwd = gwdirb; *cwd = '\0';
#ifdef BSD
	  val = getwd (gwdirb);
#else
	  val = getcwd (gwdirb, sizeof gwdirb);
#endif
	  if (val != 0)
	    {
	      strcat (cwd, "/");
	    }
	  else
	    {
	      fprintf (stderr, cwd);
	      *cwd = '\0';
	    }
	  strcat (msgp->mtext, cwd);
	}
	
      strcat (msgp->mtext, argv[0]);
      strcat (msgp->mtext, " ");
      argv++; argc--;
    }
  strcat (msgp->mtext, "\n");
  msgp->mtype = 1;
  if (msgsnd (s, msgp, strlen (msgp->mtext)+1, 1, 0) < 0)
    {
      perror ("msgsnd");
      exit (1);
    }
  /*
   * Now, wait for an answer
   */
  printf ("Waiting\n");

  msgrcv (s, msgp, BUFSIZ, getpid (), 0);	/* wait for anything back */
  strcpy (buf, msgp->mtext);

  printf ("Got back: %s\n", buf);
  exit (0);
}

#endif /* QNX */

#endif /* IPC */

#endif /* BSD && IPC */
