;;;;Keyboard initialization file for the QNX console.
;;;
;;; Written by Walker D. Tingley, July 10, 1992

(require 'keypad)			;We map through the keypad package

;;; In Berkeley Unix, at least of the type we have (Ultrix V4.1),
;;; We must use interrupt input to get an eight bit path.
;;; This is done by default on the standard Berkeley Unix configuration
;;; and for it, this code is not needed.
;;; I have set the default to use Cbreak mode as I consider the interrupt input
;;; mode to be potentially buggy.  This is ok for normal terminals, but the
;;; QNX console function keys send 8-bit sequences, so we MUST have an 8-bit
;;; transparent path.
;;; If the interrupt mode is allready set, this code is essentially a NOP.

(if (eq system-type 'berkeley-unix)	;You may bag this if you enable
    (set-input-mode t nil))		;INTERRUPT_INPUT in config.c

(defvar QNX-map nil
  "The QNX-map maps the function keys on the QNX console.")

(if (not QNX-map)			;If .emacs hasn't allready set it up
    (progn
      (setq QNX-map (lookup-key global-map "\e\177\e"))
      (if (not (keymapp QNX-map))
	  (progn
	    (define-key global-map "\e\177" nil) ; Allow \e\177 to be a prefix
	    (setq QNX-map (make-keymap)))) ; <ESC> DEL <ESC> commands
      (setup-terminal-keymap QNX-map
			     '(("\001" . 1) ;Unshifted function keys F1-F12
			       ("\002" . 2)
			       ("\003" . 3)
			       ("\004" . 4)
			       ("\005" . 5)
			       ("\006" . 6)
			       ("\007" . 7)
			       ("\010" . 8)
			       ("\011" . 9)
			       ("\012" . 10)
			       ("\056" . 11)
			       ("\057" . 12)
			       ("\013" . 13) ;Shifted F1-F12 ==> F13-F24
			       ("\014" . 14)
			       ("\015" . 15)
			       ("\016" . 16)
			       ("\017" . 17)
			       ("\020" . 18)
			       ("\021" . 19)
			       ("\022" . 20)
			       ("\023" . 21)
			       ("\024" . 22)
			       ("\133" . 23)
			       ("\134" . 24)
			       ("\040" . ?h) ;Home
			       ("\041" . ?u) ;Up
			       ("\042" . ?P) ;Page-up ==> Previous page
			       ("\043" . ?-) ;Keypad -
			       ("\044" . ?l) ;Left
			       ("\045" . ?5) ;Keypad 5
			       ("\046" . ?r) ;Right
			       ("\047" . ?,) ;Plus ==> Keypad ,
;			       ("\050" . ??) ;End ==> Unassigned
			       ("\051" . ?d) ;Down
			       ("\052" . ?N) ;Page-down ==> Next page
			       ("\053" . ?I) ;Insert ==> Insert character
			       ("\054" . ?D) ;Keypad Del ==> Delete character
			       ("\060" . ?7) ;Map Ctrl keypad 0-9 to Keypad 0-9
			       ("\061" . ?8)
			       ("\062" . ?9)
			       ("\064" . ?4)
			       ("\065" . ?5)
			       ("\066" . ?6)
			       ("\070" . ?1)
			       ("\071" . ?2)
			       ("\072" . ?3)
			       ("\073" . ?0)
			       ("\074" . ?.) ;Ctrl keypad . ==> Keypad .
			       ("\120" . ?e))) ;Ctrl Enter . ==> Keypad Enter
      (define-key global-map "\e\177\e" QNX-map)))
