#include "config.h"
#include <stdio.h>

/* fcntl.h included in qnx_map.h */

extern int qnx_first_data, qnx_first_bss, qnx_last_data, qnx_last_bss;
extern int qnx_out_initial;

/*
This file dumps and restores the _DATA and _BSS sections of the program that
are used by the program.  Anything else is either constant data or belongs to
a library function and is not touched.

At the times save_data and restore_data are used, it is safe to just exit
when we get an error.
*/

static fatalsys(str)
char *str;
{
  perror(str);
  exit(1);
}

static fatalerr(str)
char *str;
{
  fputs(str, stderr);
  putc('\n', stderr);
  exit(1);
}

static safe_read(fd, buf, cnt)
int fd, cnt;
char *buf;
{
  int rtn=read(fd, buf, cnt);
  if (rtn != cnt) {
    if (rtn < 0)
      fatalsys("read");
    else
      fatalerr("short count in read");
  }
  return(rtn);
}

static safe_write(fd, buf, cnt)
int fd, cnt;
char *buf;
{
  int rtn=write(fd, buf, cnt);
  if (rtn != cnt) {
    if (rtn < 0)
      fatalsys("write");
    else
      fatalerr("short count in write");
  }
  return(rtn);
}

struct dump_hdr {
  char *dataloc;
  int datasize;
  char *bssloc;
  int bsssize;
};

save_data(file)
char *file;
{
  struct dump_hdr hdr;
  int fd;
  if(qnx_out_initial)
    fatalerr(
"Out of initial allocation. Need more memory (QNX_ALLOCATION_SIZE).");
  fd=open(file, O_WRONLY|O_CREAT, 0666);
  if(fd < 0)
    fatalsys("open");
  hdr.dataloc = ((char *) &qnx_first_data);
  hdr.datasize = ((char *) &qnx_last_data) - ((char *) &qnx_first_data);
  hdr.bssloc = ((char *) &qnx_first_bss);
  hdr.bsssize = ((char *) &qnx_last_bss) - ((char *) &qnx_first_bss);
  safe_write(fd, &hdr, sizeof(hdr));
  safe_write(fd, hdr.dataloc, hdr.datasize);
  safe_write(fd, hdr.bssloc, hdr.bsssize);
  close(fd);
}

restore_data(file)
char *file;
{
  struct dump_hdr hdr;
  int fd=open(file, O_RDONLY);
  if(fd < 0)
    fatalsys("open");
  safe_read(fd, &hdr, sizeof(hdr));
  /* Consistancy checks */
  if (hdr.dataloc != ((char *) &qnx_first_data) ||
      hdr.datasize != ((char *) &qnx_last_data) - ((char *) &qnx_first_data) ||
      hdr.bssloc != ((char *) &qnx_first_bss) ||
      hdr.bsssize != ((char *) &qnx_last_bss) - ((char *) &qnx_first_bss))
    fatalerr("Inconsistancy between dump file and program");
  safe_read(fd, hdr.dataloc, hdr.datasize);
  safe_read(fd, hdr.bssloc, hdr.bsssize);
  close(fd);
}
