/* System header for QNX */

#define CANNOT_DUMP 1

/*
 *	Define symbols to identify the version of Unix this is.
 *	Define all the symbols that apply correctly.
 */

#define QNX

/* SYSTEM_TYPE should indicate the kind of system you are using.
 It sets the Lisp variable system-type.  */

/*
While not generally documented, it is standard to use lower case for this.
All other s-*.h files do this.
*/

#define SYSTEM_TYPE "qnx"

/* NOMULTIPLEJOBS should be defined if your system's shell
 does not have "job control" (the ability to stop a program,
 run some other program, then continue the first one).  */

#define NOMULTIPLEJOBS

/* Emacs can read input using SIGIO and buffering characters itself,
   or using CBREAK mode and making C-g cause SIGINT.
   The choice is controlled by the variable interrupt_input.
   Define INTERRUPT_INPUT to make interrupt_input = 1 the default (use SIGIO)

   SIGIO can be used only on systems that implement it (4.2 and 4.3).
   CBREAK mode has two disadvatages
     1) At least in 4.2, it is impossible to handle the Meta key properly.
        I hear that in system V this problem does not exist.
     2) Control-G causes output to be discarded.
        I do not know whether this can be fixed in system V.

   Another method of doing input is planned but not implemented.
   It would have Emacs fork off a separate process
   to read the input and send it to the true Emacs process
   through a pipe.
*/

/* #define INTERRUPT_INPUT */

/* Letter to use in finding device name of first pty,
  if system supports pty's.  'a' means it is /dev/ptya0  */

#define FIRST_PTY_LETTER 'p'

/*
 *	Define HAVE_TIMEVAL if the system supports the BSD style clock values.
 *	Look in <sys/time.h> for a timeval structure.
 */

#define HAVE_TIMEVAL

/*
 *	Define HAVE_SELECT if the system supports the `select' system call.
 */

#define HAVE_SELECT

/*
 *	Define HAVE_PTYS if the system supports pty devices.
 */

#define HAVE_PTYS

/*
 *	Define NONSYSTEM_DIR_LIBRARY to make Emacs emulate
 *      The 4.2 opendir, etc., library functions.
 */

/* #define NONSYSTEM_DIR_LIBRARY */

/* Define this symbol if your system has the functions bcopy, etc. */

/* #define BSTRING */

/* subprocesses should be defined if you want to
   have code for asynchronous subprocesses
   (as used in M-x compile and M-x shell).
   This is generally OS dependent, and not supported
   under most USG systems. */

#define subprocesses

/* If your system uses COFF (Common Object File Format) then define the
   preprocessor symbol "COFF". */

/* #define COFF */

/* define MAIL_USE_FLOCK if the mailer uses flock
   to interlock access to /usr/spool/mail/$USER.
   The alternative is that a lock file named
   /usr/spool/mail/$USER.lock.  */

/* #define MAIL_USE_FLOCK */

/* Define CLASH_DETECTION if you want lock files to be written
   so that Emacs can tell instantly when you try to modify
   a file that someone else has modified in his Emacs.  */

#define CLASH_DETECTION

/* Here, on a separate page, add any special hacks needed
   to make Emacs work on this system.  For example,
   you might define certain system call names that don't
   exist on your system, or that do different things on
   your system and must be used only through an encapsulation
   (Which you should place, by convention, in sysdep.c).  */

/*
Include files declaring functions taking a variable number of arguments
or that are redefined below.  All of these are protected from double
inclusion, so if they are included within the source file, that is harmless.
This insures that the redefined functions are declared consistantly within
emacs.
*/

#ifdef emacs
#include <unistd.h>
#include <stdlib.h>
#include <malloc.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#undef max
#undef min
/*
Anything that needs the conventional meaning of NULL includes stdio.h
afterwards.
*/
#undef NULL
#endif

#define SYSV_SYSTEM_DIR
#define HAVE_CLOSEDIR

/*
Allow our allocator and the qnx allocator to coexist in memory.  Let libraries
use their version and we'll use our version.
We can bag this if we discover later that we can completly replace
the QNX malloc with our own.
*/

#ifdef emacs			/* Don't do this for xmakefile or ../etc */

#define malloc sys_malloc
#define realloc sys_realloc
#define free sys_free
#define sbrk sys_sbrk

/*
alloca is defined using NULL, which is used as something else, so we must not
use it.  This definition raises an error instead of returning NULL if we
run out of stack space.  This is what emacs expects.

The old definition which this modification is based on is in <stdlib.h> and is:

#define alloca( s )     ((__ALLOCA_ALIGN(s)<stackavail())?__alloca(s):NULL)

If this definition has changed in your current version of QNX, then change
the definition below to correspond.
*/

void *out_of_stack(void);

#undef alloca
#define alloca(size) \
( (__ALLOCA_ALIGN(size) < stackavail()) ? __alloca(size) : out_of_stack())

/*
abort is virtually useless as a debugger breakpoint, as there is no way to
get a backtrace from that point.
*/

void sys_abort(char *file, int line);

#define abort() sys_abort(__FILE__, __LINE__)

/*
Certain calls are treated as if they don't interrupt, but they do here.
Make emacs call special versions which recover from interrupts.
Use longs for character counts.
*/

#define read sys_read
#define write sys_write
#define open sys_open
#define close sys_close

#define INTERRUPTABLE_OPEN
#define INTERRUPTABLE_CLOSE
#define INTERRUPTABLE_IO

/*
These emacs functions take variable arguments and must be declared as such
with prototypes to compile without errors, and to be correctly called with
stack arguments.
*/

extern void fatal(char *str, ...);
extern void error (char *m, ...);
extern int sys_open(char *path, int oflag, ...);
extern char *tparam(char *string, char *outstring, int len, ...);
extern void message(char *m, ...);

/* Redefine BSD string functions as their POSIX equivalents */

#define index strchr
#define rindex strrchr

/*
Redefine XTYPE and XGCTYPE to not cast to enum Lisp_Type.
This gets around a WATCOM C bug:

((enum Lisp_Type) ((a) >> VALBITS)) puts a in an 8-bit register, then
shifts it.  This should be done the other way around.
*/


#define XTYPE(a) ((int) ((a) >> VALBITS))
#define XGCTYPE(a) ((int) (((a) >> VALBITS) & GCTYPEMASK))

#endif /* emacs */

/* longjmp does what we want; setjmp is defined as _setjmp */

#define _longjmp longjmp

/* Some compilers tend to put everything declared static
   into the initialized data area, which becomes pure after dumping Emacs.
   On these systems, you must #define static as nothing to foil this.
   Note that emacs carefully avoids static vars inside functions.  */

/* #define static */

/* No support for SIGTSTP even though it is defined */

#undef SIGTSTP

/* No support for SIGIO even though it is defined */

#undef SIGIO

/* Makefile stuff */

#define LD_SWITCH_SYSTEM -N 256k -g -M "-Wl,option verbose"
#define ld cc
