/* Interfaces to system-dependent kernel and library entries.
   Copyright (C) 1985, 1986, 1987, 1988 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.  No author or distributor
accepts responsibility to anyone for the consequences of using it
or for whether it serves any particular purpose or works at all,
unless he says so in writing.  Refer to the GNU Emacs General Public
License for full details.

Everyone is granted permission to copy, modify and redistribute
GNU Emacs, but only under the conditions described in the
GNU Emacs General Public License.   A copy of this license is
supposed to have been given to you along with GNU Emacs so you
can know your rights and responsibilities.  It should be in a
file named COPYING.  Among other things, the copyright notice
and this notice must be preserved on all copies.  */


#include <signal.h>
#include <setjmp.h>

#include "config.h"
#include "lisp.h"
#undef NULL

#define min(x,y) ((x) > (y) ? (y) : (x))

/* In this file, open, read and write refer to the system calls,
   not our sugared interfaces  sys_open, sys_read and sys_write.
   Contrariwise, for systems where we use the system calls directly,
   define sys_read, etc. here as aliases for them.  */
#ifndef read
#define sys_read read
#define sys_write write
#endif /* `read' is not a macro */

#undef read
#undef write

#ifndef close
#define sys_close close
#else 
#undef close
#endif

#ifndef open
#define sys_open open
#else /* `open' is a macro */
#undef open
#endif /* `open' is a macro */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

extern int errno;
#ifndef QNX
extern char *sys_errlist[];
#else
#undef abort				/* use real abort here */
#include <sys/types.h>
#include <sys/utsname.h>
/* #include <string.h> */
#include <termios.h>
#include <process.h>
#include <sys/wait.h>
#include <time.h>
#include <sys/time.h>
#define MAXPATHLEN 1024			/* To agree with what buffer.c uses */
#endif

#ifndef BSD4_1
#ifdef BSD /* this is done this way to avoid defined(BSD) || defined (USG)
	      because the vms compiler doesn't grok `defined' */
#include <fcntl.h>
#endif
#ifdef USG
#include <fcntl.h>
#endif
#endif /* not 4.1 bsd */

#ifdef BSD
#include <sys/ioctl.h>
#ifdef BSD4_1
#include <wait.h>
#else /* not 4.1 */
#include <sys/wait.h>
#endif /* not 4.1 */
#endif /* BSD */

#ifdef STRIDE
#include <sys/ioctl.h>
#endif 

#ifdef mips
#include <sys/ioctl.h>
#endif 

/* Get rid of LLITOUT in 4.1, since it is said to stimulate kernel bugs.  */
#ifdef BSD4_1
#undef LLITOUT
#define LLITOUT 0
#endif /* 4.1 */

#ifdef HAVE_TERMIO
#include <termio.h>
#undef TIOCGETP
#define TIOCGETP TCGETA
#undef TIOCSETN
#define TIOCSETN TCSETAW
#undef TIOCSETP
#define TIOCSETP TCSETAF
#define TERMINAL struct termio
#define OSPEED(str) (str.c_cflag & CBAUD)
#define SETOSPEED(str,new) (str.c_cflag = (str.c_cflag & ~CBAUD) | (new))
#define TABS_OK(str) ((str.c_oflag & TABDLY) != TAB3)
#endif /* HAVE_TERMIO */

#ifdef XENIX
#undef TIOCGETC  /* Avoid confusing some conditionals that test this.  */
#endif

#ifndef HAVE_TERMIO
#ifndef QNX
#include <sgtty.h>
#define TERMINAL struct sgttyb
#define OSPEED(str) str.sg_ospeed
#define SETOSPEED(str,new) (str.sg_ospeed = (new))
#define TABS_OK(str) ((str.sg_flags & XTABS) != XTABS)
#undef TCSETAW
#define TCSETAW TIOCSETN
#endif /* not QNX */
#endif /* not HAVE_TERMIO */

#ifdef USG
#include <sys/utsname.h>
#include <memory.h>
#include <string.h>
#ifdef TIOCGWINSZ
#ifndef IRIS_4D
#ifndef mips
/* Some USG systems with TIOCGWINSZ need this file; some don't have it.
   We don't know how to distinguish them.
   If this #include gets an error, just delete it.  */
#include <sys/sioctl.h>
#endif
#endif
#endif
#ifdef HAVE_TIMEVAL
#ifdef HPUX
#include <time.h>
#else
#include <sys/time.h>
#endif
#endif /* HAVE_TIMEVAL */
#endif /* USG */

#ifdef NEED_BSDTTY
#include <sys/bsdtty.h>
#endif 

#include "termhooks.h"
#include "termchar.h"
#include "termopts.h"
#include "dispextern.h"

#ifdef NONSYSTEM_DIR_LIBRARY
#include "ndir.h"
#endif /* NONSYSTEM_DIR_LIBRARY */

#ifndef sigmask
#define sigmask(no) (1L << ((no) - 1))
#endif

/* Define SIGCHLD as an alias for SIGCLD.  There are many conditionals
   testing SIGCHLD.  */

#ifdef SIGCLD
#ifndef SIGCHLD
#define SIGCHLD SIGCLD
#endif /* not SIGCHLD */
#endif /* SIGCLD */

#ifndef QNX
static int baud_convert[] =
#ifdef BAUD_CONVERT
  BAUD_CONVERT;
#else
  {
    0, 50, 75, 110, 135, 150, 200, 300, 600, 1200,
    1800, 2400, 4800, 9600, 19200, 38400
  };
#endif
#endif

#ifndef QNX
extern short ospeed;
#else
extern long ospeed;		/* QNX stores real baud rate here */
#endif

discard_tty_input ()
{
#ifndef QNX
  TERMINAL buf;
#else
  /* char buf[16]; */
#endif

  if (noninteractive)
    return;

#ifndef QNX
  ioctl (0, TIOCGETP, &buf);
  ioctl (0, TIOCSETP, &buf);
#else
  /* while (dev_read(0, buf, sizeof(buf), 0, 0, 0, 0, 0) > 0); */
  tcflush(0, TCIFLUSH);
#endif
}

#ifdef SIGTSTP

stuff_char (c)
     char c;
{
/* Should perhaps error if in batch mode */
#ifdef TIOCSTI
  ioctl (0, TIOCSTI, &c);
#else /* no TIOCSTI */
  error ("Cannot stuff terminal input characters in this version of Unix.");
#endif /* no TIOCSTI */
}

#endif /* SIGTSTP */

init_baud_rate ()
{
#ifndef QNX
  TERMINAL sg;

  if (noninteractive)
    ospeed = 0;
  else
    {
      SETOSPEED (sg, B9600);
      ioctl (0, TIOCGETP, &sg);
      ospeed = OSPEED (sg);
    }
  baud_rate = ospeed == 0 ? 1200
    : ospeed < sizeof baud_convert / sizeof baud_convert[0]
      ? baud_convert[ospeed] : 9600;
#else
  struct termios term;
  if (noninteractive)
    ospeed=9600;
  else
    {
      term.c_ospeed = 9600;		/* in case we fail */
      tcgetattr(0, &term);
      ospeed = term.c_ospeed;
    }
  baud_rate = ospeed;
#endif
}

/*ARGSUSED*/
set_exclusive_use (fd)
     int fd;
{
#ifdef QNX
  fcntl(fd, F_SETFD, FD_CLOEXEC); /* may be usable on other systems too */
#else
#ifdef FIOCLEX
  ioctl (fd, FIOCLEX, 0);
#endif
  /* Ok to do nothing if this feature does not exist */
#endif /* QNX */
}

#ifndef subprocesses

wait_without_blocking ()
{
#ifdef BSD
  wait3 (0, WNOHANG | WUNTRACED, 0);
#else
#ifndef QNX
  croak ("wait_without_blocking");
#else
  waitpid(-1, 0, WNOHANG);
#endif
#endif
}

#endif /* not subprocesses */

int wait_debugging;   /* Set nonzero to make following function work under dbx
		         (at least for bsd).  */

/* Wait for subprocess with process id `pid' to terminate and
   make sure it will get eliminated (not remain forever as a zombie) */

wait_for_termination (pid)
     int pid;
{
  while (1)
    {
#ifdef subprocesses
#if defined(BSD) || (defined(HPUX) && !defined(HPUX_5))
      /* Note that kill returns -1 even if the process is just a zombie now.
	 But inevitably a SIGCHLD interrupt should be generated
	 and child_sig will do wait3 and make the process go away. */
      /* There is some indication that there is a bug involved with
	 termination of subprocesses, perhaps involving a kernel bug too,
	 but no idea what it is.  Just as a hunch we signal SIGCHLD to see
	 if that causes the problem to go away or get worse.  */
#ifdef BSD4_1
      extern int synch_process_pid;
      sighold (SIGCHLD);
      if (synch_process_pid == 0)
	{
          sigrelse (SIGCHLD);
	  break;
	}
      if (wait_debugging)
	sleep (1);
      else
	sigpause (SIGCHLD);
#else /* not BSD4_1 */
      sigsetmask (1 << (SIGCHLD - 1));
      if (0 > kill (pid, 0))
        {
	  sigsetmask (0);
	  kill (getpid (), SIGCHLD);
	  break;
	}
      if (wait_debugging)
	sleep (1);
      else
	sigpause (0);
#endif /* not BSD4_1 */
#else /* not BSD, and not HPUX version >= 6 */
#ifdef UNIPLUS
      if (0 > kill (pid, 0))
	break;
      wait (0);
#else /* neither BSD nor UNIPLUS: random sysV */
      if (0 > kill (pid, 0))
	break;
      pause ();
#endif /* not UNIPLUS */
#endif /* not BSD, and not HPUX version >= 6 */
#else /* not subprocesses */
#ifndef BSD4_1
      if (0 > kill (pid, 0))
	break;
      wait (0);
#else /* BSD4_1 */
      int status;
      status = wait (0);
      if (status == pid || status == -1)
	break;
#endif /* BSD4_1 */
#endif /* not subprocesses */
    }
}

#ifdef subprocesses

/*
 *	flush any pending output
 *      (may flush input as well; it does not matter the way we use it)
 */
 
flush_pending_output (channel)
     int channel;
{
#ifdef TCFLSH
  ioctl (channel, TCFLSH, 1);
#else
#ifdef TIOCFLUSH
  int zero = 0;
  /* 3rd arg should be ignored
     but some 4.2 kernels actually want the address of an int
     and nonzero means something different.  */
  ioctl (channel, TIOCFLUSH, &zero);
#endif
#endif
}

/*  Set up the terminal at the other end of a pseudo-terminal that
    we will be controlling an inferior through.
    It should not echo or do line-editing, since that is done
    in Emacs.  No padding needed for insertion into an Emacs buffer.  */

child_setup_tty (out)
     int out;
{
#ifdef QNX
  struct termios term;
  long disable = fpathconf(0, _PC_VDISABLE);
  if(tcgetattr(out, &term) < 0)	/* not a tty */
    return;
  term.c_oflag &= ~OPOST;	/* There is no ONLCR, so we do it this way */
  term.c_lflag &= ~ECHO;
  term.c_lflag |= ISIG|ICANON;
  term.c_cc[VEOF] = 04;
  term.c_cc[VERASE] = disable;
  term.c_cc[VKILL] = disable;
  tcsetattr(out, TCSADRAIN, &term);
#else
  TERMINAL s;

  ioctl (out, TIOCGETP, &s);
#ifdef HAVE_TERMIO
  s.c_oflag |= OPOST;		/* Enable output postprocessing */
  s.c_oflag &= ~ONLCR;		/* Disable map of NL to CR-NL on output */
  s.c_oflag &= ~(NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);	/* No output delays */
  s.c_lflag &= ~ECHO;		/* Disable echo */
  s.c_lflag |= ISIG;		/* Enable signals */
  s.c_iflag &= ~IUCLC;		/* Disable map of upper case to lower on input */
  s.c_oflag &= ~OLCUC;		/* Disable map of lower case to upper on output */
/* said to be unnecesary
  s.c_cc[VMIN] = 1;		/* minimum number of characters to accept
  s.c_cc[VTIME] = 0;		/* wait forever for at least 1 character
*/
  s.c_lflag |= ICANON;		/* Enable erase/kill and eof processing */
  s.c_cc[VEOF] = 04;		/* insure that EOF is Control-D */
  s.c_cc[VERASE] = 0377;	/* disable erase processing */
  s.c_cc[VKILL] = 0377;		/* disable kill processing */
#ifdef HPUX
  s.c_cflag = (s.c_cflag & ~CBAUD) | B9600; /* baud rate sanity */
#endif HPUX
#ifdef IBMRTAIX
/* AIX enhanced edit loses NULs, so disable it */
  s.c_line = 0;
  s.c_iflag &= ~ASCEDIT;
  /* Also, PTY overloads NUL and BREAK.
     don't ignore break, but don't signal either, so it looks like NUL.  */
  s.c_iflag &= ~IGNBRK;
  s.c_iflag &= ~BRKINT;
/* QUIT and INTR work better as signals, so disable character forms */
  s.c_cc[VQUIT] = 0377;
  s.c_cc[VINTR] = 0377;
  s.c_cc[VEOL] = 0377;
  s.c_lflag &= ~ISIG;
  s.c_cflag = (s.c_cflag & ~CBAUD) | B9600; /* baud rate sanity */
#endif /* IBMRTAIX */

#else /* not HAVE_TERMIO */
  s.sg_flags &= ~(ECHO | CRMOD | ANYP | ALLDELAY | RAW | LCASE | CBREAK | TANDEM);
#endif /* not HAVE_TERMIO */

  ioctl (out, TIOCSETN, &s);

#ifdef BSD4_1
  if (interrupt_input)
    reset_sigio ();
#endif /* BSD4_1 */
#ifdef RTU
  {
    int zero = 0;
    ioctl (out, FIOASYNC, &zero);
  }
#endif /* RTU */
#endif /* QNX */
}

#endif /* subprocesses */

/*ARGSUSED*/
setpgrp_of_tty (pid)
     int pid;
{
#ifdef TIOCSPGRP
  ioctl (0, TIOCSPGRP, &pid);
#else
#ifdef QNX
  tcsetct(0, pid);
  tcsetpgrp(0, pid);
#else
  /* Just ignore this for now and hope for the best */
#endif
#endif
}

/* Suspend the Emacs process; give terminal to its superior.  */

sys_suspend ()
{
#ifdef SIGTSTP
#ifdef BSD
  killpg (getpgrp (0), SIGTSTP);
#else
  kill (-getpgrp (0), SIGTSTP);
#endif

#else
#ifdef USG_JOBCTRL /* If you don't know what this is don't mess with it */
  ptrace (0, 0, 0, 0);		/* set for ptrace - caught by csh */
  kill (getpid (), SIGQUIT);

#else

/* On a system where suspending is not implemented,
   instead fork a subshell and let it talk directly to the terminal
   while we wait.  */
#ifdef QNX
  char previous_dir[1024];
  int pid;
  int (*inter) ();
  int (*quit) ();
  int (*term) ();
  Lisp_Object dir;
  unsigned char *str;
  int len;
  char *sh;
  if(!getcwd(previous_dir, 1024)) /* Save before changing */
    previous_dir[0]='\0';
  sh = (char *) egetenv ("SHELL");
  if (sh == 0)
    sh = "sh";
  /* Use our buffer's default directory for the subshell. */
  dir = intern ("default-directory");
  if (XFASTINT (Fboundp (dir)) == XFASTINT (Qnil))
    goto xyzzy;
  dir = Fsymbol_value (dir);
  if (XTYPE (dir) != Lisp_String)
    goto xyzzy;

  str = (unsigned char *) alloca (XSTRING (dir)->size + 2);
  len = XSTRING (dir)->size;
  bcopy (XSTRING (dir)->data, str, len);
  if (str[len - 1] != '/') str[len++] = '/'; /* Is this really needed ? */
  str[len] = 0;
  chdir (str);
xyzzy:
  pid=spawnlp(P_NOWAIT, sh, sh, 0);
  if(previous_dir[0])
    chdir (previous_dir);
  if (pid == -1)
    error ("Can't spawn subshell");
#else /* not QNX */
  int pid = fork ();
  int (*inter) ();
  int (*quit) ();
  int (*term) ();
#ifdef SIGIO
  int (*sigio) ();
#endif SIGIO

  if (pid == -1)
    error ("Can't spawn subshell");
  if (pid == 0)
    {
      char *sh;

      sh = (char *) egetenv ("SHELL");
      if (sh == 0)
	sh = "sh";
      /* Use our buffer's default directory for the subshell.  */
      {
	Lisp_Object dir;
	unsigned char *str;
	int len;

	/* mentioning bf_cur->buffer would mean including buffer.h,
	   which somehow wedges the hp compiler.  So instead... */

	dir = intern ("default-directory");
	/* Can't use NULL */
	if (XFASTINT (Fboundp (dir)) == XFASTINT (Qnil))
	  goto xyzzy;
	dir = Fsymbol_value (dir);
	if (XTYPE (dir) != Lisp_String)
	  goto xyzzy;

	str = (unsigned char *) alloca (XSTRING (dir)->size + 2);
	len = XSTRING (dir)->size;
	bcopy (XSTRING (dir)->data, str, len);
	if (str[len - 1] != '/') str[len++] = '/';
	str[len] = 0;
	chdir (str);
      }
    xyzzy:
#ifdef subprocesses
      close_process_descs ();	/* Close Emacs's pipes/ptys */
#endif
      execlp (sh, sh, 0);
      write (1, "Can't execute subshell", 22);
      _exit (1);
    }
#endif /* not QNX */
  inter = (int (*)()) signal (SIGINT, SIG_IGN);
  quit = (int (*)()) signal (SIGQUIT, SIG_IGN);
  term = (int (*)()) signal (SIGTERM, SIG_IGN);
#ifdef SIGIO
  sigio = (int (*)()) signal (SIGIO, SIG_IGN);
#endif /* SIGIO */
  wait_for_termination (pid);
  signal (SIGINT, inter);
  signal (SIGQUIT, quit);
  signal (SIGTERM, term);
#ifdef SIGIO
  signal (SIGIO, sigio);
#endif /* SIGIO */
#endif /* no USG_JOBCTRL */
#endif /* no SIGTSTP */
}

#ifdef F_SETFL

int old_fcntl_flags;

init_sigio ()
{
#ifdef FASYNC
  old_fcntl_flags = fcntl (0, F_GETFL, 0) & ~FASYNC;
#endif
  request_sigio ();
}

reset_sigio ()
{
  unrequest_sigio ();
}

#ifdef FASYNC		/* F_SETFL does not imply existance of FASYNC */

request_sigio ()
{
#ifdef SIGWINCH
  int omask = sigblock (0);
  sigsetmask (omask & ~sigmask (SIGWINCH));
#endif
  fcntl (0, F_SETFL, old_fcntl_flags | FASYNC);

  interrupts_deferred = 0;
}

unrequest_sigio ()
{
#ifdef SIGWINCH
  sigblock (sigmask (SIGWINCH));
#endif
  fcntl (0, F_SETFL, old_fcntl_flags);
  interrupts_deferred = 1;
}

#else /* no FASYNC */
#ifdef STRIDE		/* Stride doesn't have FASYNC - use FIOASYNC */

request_sigio ()
{
  int on = 1;
  ioctl (0, FIOASYNC, &on);
  interrupts_deferred = 0;
}

unrequest_sigio ()
{
  int off = 0;

  ioctl (0, FIOASYNC, &off);
  interrupts_deferred = 1;
}

#else /* not FASYNC, not STRIDE */
 
request_sigio ()
{
  croak ("request_sigio");
}
 
unrequest_sigio ()
{
  croak ("unrequest_sigio");
}
 
#endif /* STRIDE */
#endif /* FASYNC */
#endif /* F_SETFL */

#ifdef QNX
struct termios old_term;
#else
TERMINAL old_gtty;		/* The initial tty mode bits */
#endif

int term_initted;		/* 1 if outer tty status has been recorded */

#ifdef F_SETOWN
int old_fcntl_owner;
#endif /* F_SETOWN */

#ifdef TIOCGLTC
struct ltchars old_ltchars;
#endif /* TIOCGLTC */

#ifdef TIOCGETC
struct tchars old_tchars;
int old_lmode;

int lmode;			/* Current lmode value. */
				/* Needed as global for 4.1 */
#endif /* TIOCGETC */

/* This may also be defined in stdio,
   but if so, this does no harm,
   and using the same name avoids wasting the other one's space.  */

#ifdef USG
unsigned char _sobuf[BUFSIZ+8];
#else
char _sobuf[BUFSIZ];
#endif
 
#ifdef TIOCGLTC
static struct ltchars new_ltchars = {-1,-1,-1,-1,-1,-1};
#endif
#ifdef TIOCGETC
  static struct tchars new_tchars = {-1,-1,-1,-1,-1,-1};
#endif 

init_sys_modes ()
{
#ifdef QNX
  struct termios new_term;
  long disable = fpathconf(0, _PC_VDISABLE);
  if (noninteractive)
    return;
  if (!read_socket_hook) {
    tcgetattr(0, &old_term);
    new_term=old_term;
    new_term.c_iflag |= IGNBRK;
    new_term.c_iflag &= ~(ICRNL|ISTRIP);
    new_term.c_lflag &= ~(ECHO|ICANON);
    new_term.c_lflag |= ISIG;
    if (flow_control)
      new_term.c_iflag |= IXON;
    else
      new_term.c_iflag &= ~IXON;
    new_term.c_oflag &= ~OPOST;
    new_term.c_cflag |= CS8;
    new_term.c_cflag &= ~PARENB;
    new_term.c_cc[VINTR] = '\007';
    new_term.c_cc[VQUIT] = disable;
    new_term.c_cc[VMIN] = 1;
    new_term.c_cc[VTIME] = 0;
    tcsetattr(0, TCSADRAIN, &new_term);
  }
#else
  TERMINAL sg;
#ifdef TIOCGETC
  struct tchars tchars;
#endif

  if (noninteractive)
    return;

  ioctl (0, TIOCGETP, &old_gtty);
  if (!read_socket_hook)
    {
      sg = old_gtty;

#ifdef HAVE_TERMIO
      sg.c_iflag |= (IGNBRK);	/* Ignore break condition */
      sg.c_iflag &= ~ICRNL;	/* Disable map of CR to NL on input */
#ifdef ISTRIP
      sg.c_iflag &= ~ISTRIP;	/* don't strip 8th bit on input */
#endif
      sg.c_lflag &= ~ECHO;	/* Disable echo */
      sg.c_lflag &= ~ICANON;	/* Disable erase/kill processing */
      sg.c_lflag |= ISIG;	/* Enable signals */
      if (flow_control)
	{
	  sg.c_iflag |= IXON;	/* Enable start/stop output control */
#ifdef IXANY
	  sg.c_iflag &= ~IXANY;
#endif /* IXANY */
	}
      else
	sg.c_iflag &= ~IXON;	/* Disable start/stop output control */
      sg.c_oflag &= ~ONLCR;	/* Disable map of NL to CR-NL on output */
      sg.c_oflag &= ~TAB3;	/* Disable tab expansion */
#ifdef CS8
      sg.c_cflag |= CS8;	/* allow 8th bit on input */
      sg.c_cflag &= ~PARENB;	/* Don't check parity */
#endif
      sg.c_cc[VINTR] = '\007';	/* ^G gives SIGINT */
      /* Set up C-g for both SIGQUIT and SIGINT.
	 We don't know which we will get, but we handle both alike
	 so which one it really gives us does not matter.  */
      sg.c_cc[VQUIT] = '\007';
      sg.c_cc[VMIN] = 1;	/* Input should wait for at least 1 char */
      sg.c_cc[VTIME] = 0;	/* no matter how long that takes.  */
#ifdef VSWTCH
      sg.c_cc[VSWTCH] = CDEL;	/* Turn off shell layering use of C-z */
#endif /* VSWTCH */
#ifdef IBMRTAIX
      /* AIX enhanced edit loses NULs, so disable it */
      sg.c_line = 0;
      sg.c_iflag &= ~ASCEDIT;
      /* Also, PTY overloads NUL and BREAK.
	 don't ignore break, but don't signal either, so it looks like NUL.
	 This really serves a purpose only if running in an XTERM window
	 or via TELNET or the like, but does no harm elsewhere.  */
      sg.c_iflag &= ~IGNBRK;
      sg.c_iflag &= ~BRKINT;
#endif
#else /* if not HAVE_TERMIO */
      sg.sg_flags &= ~(ECHO | CRMOD | XTABS);
      sg.sg_flags |= ANYP;
      sg.sg_flags |= interrupt_input ? RAW : CBREAK;
#endif /* not HAVE_TERMIO */

      ioctl (0, TIOCSETN, &sg);

      /* This code added to insure that, if flow-control is not to be used,
	 we have an unlocked screen at the start. */
#ifdef TCXONC
      if (!flow_control) ioctl (0, TCXONC, 1);
#endif
#ifdef TIOCSTART
      if (!flow_control) ioctl (0, TIOCSTART, 0);
#endif

#ifdef AIX
      hft_init ();
#endif

#ifdef F_SETFL
#ifdef F_GETOWN		/* F_SETFL does not imply existance of F_GETOWN */
      if (interrupt_input)
	{
	  old_fcntl_owner = fcntl (0, F_GETOWN, 0);
	  fcntl (0, F_SETOWN, getpid ());
	  init_sigio ();
	}
#endif /* F_GETOWN */
#endif /* F_SETFL */

      /* If going to use CBREAK mode, we must request C-g to interrupt
	   and turn off start and stop chars, etc.
	   If not going to use CBREAK mode, do this anyway
	   so as to turn off local flow control for user coming over
	   network on 4.2; in this case, only t_stopc and t_startc really matter.  */
#ifdef TIOCGLTC
      ioctl (0, TIOCGLTC, &old_ltchars);
#endif /* TIOCGLTC */
#ifndef HAVE_TERMIO
#ifdef TIOCGETC
      ioctl (0, TIOCGETC, &old_tchars);
      ioctl (0, TIOCLGET, &old_lmode);

      /* Note: if not using CBREAK mode, it makes no difference how we set this */
      tchars = new_tchars;
      tchars.t_intrc = 07;
      if (flow_control)
	{
	  tchars.t_startc = '\021';
	  tchars.t_stopc = '\023';
	}
/* LPASS8 is new in 4.3, and makes cbreak mode provide all 8 bits.  */
#ifndef LPASS8
#define LPASS8 0
#endif

#ifdef BSD4_1
#define LNOFLSH 0100000
#endif

      lmode = LDECCTQ | LLITOUT | LPASS8 | LNOFLSH | old_lmode;

      ioctl (0, TIOCSETC, &tchars);
      ioctl (0, TIOCLSET, &lmode);
#endif /* TIOCGETC */
#endif /* not HAVE_TERMIO */
#ifdef TIOCGLTC
      ioctl (0, TIOCSLTC, &new_ltchars);
#endif /* TIOCGLTC */

#ifdef BSD4_1
      if (interrupt_input)
	init_sigio ();
#endif
    }
#endif /* QNX */
#ifdef _IOFBF
  /* This symbol is defined on recent USG systems.
     Someone says without this call USG won't really buffer the file
     even with a call to setbuf(). */
  setvbuf (stdout, _sobuf, _IOFBF, sizeof _sobuf);
#else
  setbuf (stdout, _sobuf);
#endif
  set_terminal_modes ();
  if (term_initted && no_redraw_on_reenter)
    {
      if (display_completed)
	direct_output_forward_char (0);
    }
  else
    screen_garbaged = 1;
  term_initted = 1;
}

/* Return nonzero if safe to use tabs in output.
   At the time this is called, init_sys_modes has not been done yet.  */
   
tabs_safe_p ()
{
#ifdef QNX
  return 1;			/* QNX has no tab expansion */
#else
  TERMINAL sg;
  if (noninteractive)
    return 1;
  ioctl (0, TIOCGETP, &sg);
  return (TABS_OK(sg));
#endif /* QNX */
}

/* Get terminal size from system.
   Store number of lines into *heightp and width into *widthp.
   If zero or a negative number is stored, the value is not valid.  */

get_screen_size (widthp, heightp)
     int *widthp, *heightp;
{
/* Define the 4.3 names in terms of the Sun names
   if the latter exist and the former do not.  */
#ifdef TIOCGSIZE
#ifndef TIOCGWINSZ
#define TIOCGWINSZ TIOCGSIZE
#define winsize ttysize
#define ws_row ts_lines
#define ws_col ts_cols
#endif
#endif /* Sun */

/* Do it using the 4.3 names if possible.  */
#ifdef TIOCGWINSZ
  struct winsize size;
  *widthp = 0;
  *heightp = 0;
  if (ioctl (0, TIOCGWINSZ, &size) < 0)
    return;
  if ((unsigned) size.ws_col > MScreenWidth
      || (unsigned) size.ws_row > MScreenLength)
    return;
  *widthp = size.ws_col;
  *heightp = size.ws_row;
#else /* not TIOCGWNSIZ */
#ifdef QNX
  unsigned int rows, cols;
  char *p;
  *widthp = 0;
  *heightp = 0;
  rows = 0;			/* Just in case */
  cols = 0;
  if (dev_size(0, -1, -1, &rows, &cols) < 0
      || rows > MScreenLength || cols > MScreenWidth) {
    rows = 0;
    cols = 0;
  }
/*
It is standard under QNX that on a serial line, environment variables
LINES and COLUMNS may be used to override the standard terminal sizes.
*/
  if(rows == 0 && (p=getenv("LINES"))) {
    rows = atoi(p);
    if (rows > MScreenLength)
      rows = 0;
  }
  if(cols == 0 && (p=getenv("COLUMNS"))) {
    cols = atoi(p);
    if (cols > MScreenWidth)
      cols = 0;
  }
  *widthp = cols;
  *heightp = rows;
#else
  *widthp = 0;
  *heightp = 0;
#endif /* QNX */
#endif /* not TIOCGWINSZ */
}

reset_sys_modes ()
{
  if (noninteractive)
    {
      fflush (stdout);
      return;
    }
  if (!term_initted)
    return;
  if (read_socket_hook)
    return;
  topos (screen_height - 1, 0);
  clear_end_of_line (screen_width);
  /* clear_end_of_line may move the cursor */
  topos (screen_height - 1, 0);
  /* Output raw CR so kernel can track the cursor hpos.  */
  cmputc ('\r');
  reset_terminal_modes ();
  fflush (stdout);
#ifdef BSD
#ifndef BSD4_1
  /* Avoid possible loss of output when changing terminal modes.  */
  fsync (fileno (stdout));
#endif
#endif
#ifdef TIOCGLTC
  ioctl (0, TIOCSLTC, &old_ltchars);
#endif /* TIOCGLTC */
#ifndef HAVE_TERMIO
#ifdef TIOCGETC
  ioctl (0, TIOCSETC, &old_tchars);
  ioctl (0, TIOCLSET, &old_lmode);
#endif /* TIOCGETC */
#endif /* not HAVE_TERMIO */
#ifdef F_SETFL
#ifdef F_SETOWN		/* F_SETFL does not imply existance of F_SETOWN */
  if (interrupt_input)
    {
      reset_sigio ();
      fcntl (0, F_SETOWN, old_fcntl_owner);
    }
#endif /* F_SETOWN */
#endif /* F_SETFL */
#ifdef BSD4_1
  if (interrupt_input)
    reset_sigio ();
#endif /* BSD4_1 */
#ifdef QNX
  while (tcsetattr(0, TCSADRAIN, &old_term) < 0 && errno == EINTR);
#else
  while (ioctl (0, TCSETAW, &old_gtty) < 0 && errno == EINTR);
#endif /* QNX */

#ifdef AIX
  hft_reset ();
#endif
}

/* Note that VMS compiler won't accept defined (CANNOT_DUMP).  */
#ifndef CANNOT_DUMP
#define NEED_STARTS
#endif

#ifndef SYSTEM_MALLOC
#ifndef NEED_STARTS
#define NEED_STARTS
#endif
#endif

#ifdef NEED_STARTS
/* Some systems that cannot dump also cannot implement these.  */

/*
 *	Return the address of the start of the text segment prior to
 *	doing an unexec().  After unexec() the return value is undefined.
 *	See crt0.c for further explanation and _start().
 *
 */

#ifndef CANNOT_DUMP
char *
start_of_text ()
{
#ifdef TEXT_START
  return ((char *) TEXT_START);
#else
#ifdef GOULD
  extern csrt();
  return ((char *) csrt);
#else /* not GOULD */
  extern int _start ();
  return ((char *) _start);
#endif /* GOULD */
#endif /* TEXT_START */
}
#endif /* not CANNOT_DUMP */

/*
 *	Return the address of the start of the data segment prior to
 *	doing an unexec().  After unexec() the return value is undefined.
 *	See crt0.c for further information and definition of data_start.
 *
 *	Apparently, on BSD systems this is etext at startup.  On
 *	USG systems (swapping) this is highly mmu dependent and
 *	is also dependent on whether or not the program is running
 *	with shared text.  Generally there is a (possibly large)
 *	gap between end of text and start of data with shared text.
 *
 *	On Uniplus+ systems with shared text, data starts at a
 *	fixed address.  Each port (from a given oem) is generally
 *	different, and the specific value of the start of data can
 *	be obtained via the UniPlus+ specific "uvar(2)" system call,
 *	however the method outlined in crt0.c seems to be more portable.
 *
 *	Probably what will have to happen when a USG unexec is available,
 *	at least on UniPlus, is temacs will have to be made unshared so
 *	that text and data are contiguous.  Then once loadup is complete,
 *	unexec will produce a shared executable where the data can be
 *	at the normal shared text boundry and the startofdata variable
 *	will be patched by unexec to the correct value.
 *
 */
 
char *
start_of_data ()
{
#ifdef DATA_START
  return ((char *) DATA_START);
#else
  extern int data_start;
  return ((char *) &data_start);
#endif
}
#endif /* NEED_STARTS (not CANNOT_DUMP or not SYSTEM_MALLOC) */

#ifndef CANNOT_DUMP
/* Some systems that cannot dump also cannot implement these.  */

/*
 *	Return the address of the end of the text segment prior to
 *	doing an unexec().  After unexec() the return value is undefined.
 */
 
char *
end_of_text ()
{
#ifdef TEXT_END
  return ((char *) TEXT_END);
#else
  extern int etext;
  return ((char *) &etext);
#endif
}
 
/*
 *	Return the address of the end of the data segment prior to
 *	doing an unexec().  After unexec() the return value is undefined.
 */

char *
end_of_data ()
{
#ifdef DATA_END
  return ((char *) DATA_END);
#else
  extern int edata;
  return ((char *) &edata);
#endif
}

#endif /* not CANNOT_DUMP */

/* Get_system_name returns as its value
 a string for the Lisp function system-name to return. */

#ifdef BSD4_1
#include <whoami.h>
#endif

#ifdef USG
/* Can't have this within the function since `static' is #defined to nothing */
static struct utsname get_system_name_name;
#endif
#ifdef QNX
static struct utsname get_system_name_name;
#endif

char *
get_system_name ()
{
#ifdef USG
  uname (&get_system_name_name);
  return (get_system_name_name.nodename);
#else /* Not USG */
#ifdef QNX
  uname (&get_system_name_name);
  return (get_system_name_name.nodename);
#else
#ifdef BSD4_1
  return sysname;
#else /* not USG, not 4.1 */
  static char system_name_saved[32];
  gethostname (system_name_saved, sizeof (system_name_saved));
  return system_name_saved;
#endif /* not USG, not 4.1 */
#endif /* not QNX */
#endif /* not USG */
}

#ifndef HAVE_SELECT

/* Emulate as much as select as is possible under 4.1 and needed by Gnu Emacs
 * Only checks read descriptors.
 */
/* How long to wait between checking fds in select */
#define SELECT_PAUSE 1
int select_alarmed;

/* For longjmp'ing back to read_input_waiting.  */

jmp_buf read_alarm_throw;

/* Nonzero if the alarm signal should throw back to read_input_waiting.
   The read_socket_hook function sets this to 1 while it is waiting.  */

int read_alarm_should_throw;

select_alarm ()
{
  select_alarmed = 1;
#ifdef BSD4_1
  sigrelse (SIGALRM);
#else /* not BSD4_1 */
  signal (SIGALRM, SIG_IGN);
#endif /* not BSD4_1 */
  if (read_alarm_should_throw)
    longjmp (read_alarm_throw, 1);
}

/* Only rfds are checked and timeout must point somewhere */
int
select (nfds, rfds, wfds, efds, timeout)
     int nfds;
     int *rfds, *wfds, *efds, *timeout;
{
  int ravail = 0, orfds = 0, old_alarm;
  extern int kbd_count;
  extern int proc_buffered_char[];
#ifndef subprocesses
  int child_changed = 0;
#else
  extern int child_changed;
#endif
  int (*old_trap) ();
  char buf;

  if (rfds)
    {
      orfds = *rfds;
      *rfds = 0;
    }
  if (wfds)
    *wfds = 0;
  if (efds)
    *efds = 0;

  /* If we are looking only for the terminal, with no timeout,
     just read it and wait -- that's more efficient.  */
  if (orfds == 1 && *timeout == 100000 && !child_changed)
    {
      if (!kbd_count)
	read_input_waiting ();
      *rfds = 1;
      return 1;
    }

  /* Once a second, till the timer expires, check all the flagged read
   * descriptors to see if any input is available.  If there is some then
   * set the corresponding bit in the return copy of rfds.
   */ 
  while (1)
    {
      register int to_check, bit, fd;

      if (rfds)
	{
	  for (to_check = nfds, bit = 1, fd = 0; --to_check >= 0; bit <<= 1, fd++)
	    {
	      if (orfds & bit)
		{
		  int avail = 0, status = 0;

		  if (bit == 1)
		    avail = detect_input_pending(); /* Special keyboard handler */
		  else
		    {
#ifdef FIONREAD
		      status = ioctl (fd, FIONREAD, &avail);
#else /* no FIONREAD */
		      /* Hoping it will return -1 if nothing available
			 or 0 if all 0 chars requested are read.  */
		      if (proc_buffered_char[fd] >= 0)
			avail = 1;
		      else
			{
			  avail = read (fd, &buf, 1);
			  if (avail > 0)
			    proc_buffered_char[fd] = buf;
			}
#endif /* no FIONREAD */
		    }
		  if (status >= 0 && avail > 0)
		    {
		      (*rfds) |= bit;
		      ravail++;
		    }
		}
	    }
	}
      if (*timeout == 0 || ravail != 0 || child_changed)
	break;
      old_alarm = alarm (0);
      old_trap = (int (*)()) signal (SIGALRM, select_alarm);
      select_alarmed = 0;
      alarm (SELECT_PAUSE);
      /* Wait for a SIGALRM (or maybe a SIGTINT) */
      while (select_alarmed == 0 && *timeout != 0 && child_changed == 0)
	{
	  /* If we are interested in terminal input,
	     wait by reading the terminal.
	     That makes instant wakeup for terminal input at least.  */
	  if (orfds & 1)
	    {
	      read_input_waiting ();
	      if (kbd_count)
		select_alarmed = 1;
	    }
	  else
	    pause();
	}
      (*timeout) -= SELECT_PAUSE;
      /* Reset the old alarm if there was one */
      alarm (0);
      signal (SIGALRM, old_trap);
      if (old_alarm != 0)
	{
	  /* Reset or forge an interrupt for the original handler. */
	  old_alarm -= SELECT_PAUSE;
	  if (old_alarm <= 0)
	    kill (getpid (), SIGALRM); /* Fake an alarm with the orig' handler */
	  else
	    alarm (old_alarm);
	}
      if (*timeout == 0)  /* Stop on timer being cleared */
	break;
    }
  return ravail;
}

/* Read keyboard input into the standard buffer,
   waiting for at least one character.  */

/* Make all keyboard buffers much bigger when using X windows.  */
#ifdef HAVE_X_WINDOWS
#define BUFFER_SIZE_FACTOR 16
#else
#define BUFFER_SIZE_FACTOR 1
#endif

read_input_waiting ()
{
  extern int kbd_count;
  extern unsigned char kbd_buffer[];
  extern unsigned char *kbd_ptr;
  int val;

  if (read_socket_hook)
    {
      read_alarm_should_throw = 0;
      if (! setjmp (read_alarm_throw))
	val = (*read_socket_hook) (0, kbd_buffer, 256 * BUFFER_SIZE_FACTOR);
      else
	val = -1;
    }
  else
    val = read (fileno (stdin), kbd_buffer, 1);

  if (val > 0)
    {
      kbd_ptr = kbd_buffer;
      kbd_count = val;
    }
}

#endif /* not HAVE_SELECT */

#ifdef BSD4_1
/* VARARGS */
setpriority ()
{
  return 0;
}

/*
 * Partially emulate 4.2 open call.
 * open is defined as this in 4.1.
 *
 * - added by Michael Bloom @ Citicorp/TTI
 *
 */

int
sys_open (path, oflag, mode)
     char *path;
     int oflag, mode;
{
  if (oflag & O_CREAT) 
    return creat (path, mode);
  else
    return open (path, oflag);
}

init_sigio ()
{
  if (noninteractive)
    return;
  lmode = LINTRUP | lmode;
  ioctl (0, TIOCLSET, &lmode);
}

reset_sigio ()
{
  if (noninteractive)
    return;
  lmode = ~LINTRUP & lmode;
  ioctl (0, TIOCLSET, &lmode);
}

request_sigio ()
{
  sigrelse (SIGTINT);

  interrupts_deferred = 0;
}

unrequest_sigio ()
{
  sighold (SIGTINT);

  interrupts_deferred = 1;
}

/* still inside #ifdef BSD4_1 */
#ifdef subprocesses

int sigheld; /* Mask of held signals */

sigholdx (signum)
     int signum;
{
  sigheld |= sigbit (signum);
  sighold (signum);
}

sigisheld (signum)
     int signum;
{
  sigheld |= sigbit (signum);
}

sigunhold (signum)
     int signum;
{
  sigheld &= ~sigbit (signum);
  sigrelse (signum);
}

sigfree ()    /* Free all held signals */
{
  int i;
  for (i = 0; i < NSIG; i++)
    if (sigheld & sigbit (i))
      sigrelse (i);
  sigheld = 0;
}

sigbit (i)
{
  return 1 << (i - 1);
}
#endif /* subprocesses */
#endif /* BSD4_1 */

#ifndef BSTRING

void
bzero (b, length)
     register char *b;
     register int length;
{
  while (length-- > 0)
    *b++ = 0;
}

/* Saying `void' requires a declaration, above, where bcopy is used
   and that declaration causes pain for systems where bcopy is a macro.  */
bcopy (b1, b2, length)
     register char *b1;
     register char *b2;
     register int length;
{
  while (length-- > 0)
    *b2++ = *b1++;
}

int
bcmp (b1, b2, length)	/* This could be a macro! */
     register char *b1;
     register char *b2;
     register int length;
{
  while (length-- > 0)
    if (*b1++ != *b2++)
      return 1;

  return 0;
}
#endif /* not BSTRING */

#ifdef BSD4_1
long random ()
{
  return (rand ());
}

srandom (arg)
     int arg;
{
  srand (arg);
}
#endif BSD4_1

#ifdef USG
/*
 *	The BSD random(3) returns numbers in the range of
 *	0 to 2e31 - 1.  The USG rand(3C) returns numbers in the
 *	range of 0 to 2e15 - 1.  This is probably not significant
 *	in this usage.
 */
  
long
random ()
{
  /* Arrange to return a range centered on zero.  */
  return rand () - (1 << 14);
}

srandom (arg)
     int arg;
{
  srand (arg);
}

#endif /* USG */

#ifdef QNX
long random ()
{
  return (rand());
}

srandom(arg)
int arg;
{
  srand (arg);
}
#endif

#ifdef INTERRUPTABLE_OPEN

int
#ifdef QNX
sys_open (char *path, int oflag, ...)
{
  int mode;
#else
/* VARARGS 2 */
sys_open (path, oflag, mode)
     char *path;
     int oflag, mode;
{
#endif
  register int rtnval;
  
#ifdef QNX
  mode = *(&oflag + 1);		/* One integer past oflag on the stack */
#endif
  while ((rtnval = open (path, oflag, mode)) == -1 && errno == EINTR);
  return (rtnval);
}

#endif /* INTERRUPTABLE_OPEN */

#ifdef INTERRUPTABLE_CLOSE

sys_close (fd)
     int fd;
{
  register int rtnval;

  while ((rtnval = close(fd)) == -1 && errno == EINTR);
  return rtnval;
}

#endif /* INTERRUPTABLE_CLOSE */

#ifdef INTERRUPTABLE_IO

int
sys_read (fildes, buf, nbyte)
     int fildes;
     char *buf;
     unsigned int nbyte;
{
  register int rtnval;
  
  while ((rtnval = read (fildes, buf, nbyte)) == -1 && errno == EINTR);
  return (rtnval);
}

int
sys_write (fildes, buf, nbyte)
     int fildes;
     char *buf;
     unsigned int nbyte;
{
  register int rtnval;

  while ((rtnval = write (fildes, buf, nbyte)) == -1 && errno == EINTR);
  return (rtnval);
}

#endif /* INTERRUPTABLE_IO */

#ifdef USG
/*
 *	All of the following are for USG.
 *
 *	On USG systems the system calls are interruptable by signals
 *	that the user program has elected to catch.  Thus the system call
 *	must be retried in these cases.  To handle this without massive
 *	changes in the source code, we remap the standard system call names
 *	to names for our own functions in sysdep.c that do the system call
 *	with retries.  Actually, for portability reasons, it is good
 *	programming practice, as this example shows, to limit all actual
 *	system calls to a single occurance in the source.  Sure, this
 *	adds an extra level of function call overhead but it is almost
 *	always negligible.   Fred Fish, Unisoft Systems Inc.
 */

char *sys_siglist[NSIG + 1] =
{
#ifdef IBMRTAIX
/* AIX has changed the signals a bit */
  "bogus signal",			/* 0 */
  "hangup",				/* 1  SIGHUP */
  "interrupt",				/* 2  SIGINT */
  "quit",				/* 3  SIGQUIT */
  "illegal instruction",		/* 4  SIGILL */
  "trace trap",				/* 5  SIGTRAP */
  "IOT instruction",			/* 6  SIGIOT */
  "crash likely",			/* 7  SIGDANGER */
  "floating point exception",		/* 8  SIGFPE */
  "kill",				/* 9  SIGKILL */
  "bus error",				/* 10 SIGBUS */
  "segmentation violation",		/* 11 SIGSEGV */
  "bad argument to system call",	/* 12 SIGSYS */
  "write on a pipe with no one to read it", /* 13 SIGPIPE */
  "alarm clock",			/* 14 SIGALRM */
  "software termination signum",	/* 15 SIGTERM */
  "user defined signal 1",		/* 16 SIGUSR1 */
  "user defined signal 2",		/* 17 SIGUSR2 */
  "death of a child",			/* 18 SIGCLD */
  "power-fail restart",			/* 19 SIGPWR */
  "bogus signal",			/* 20 */
  "bogus signal",			/* 21 */
  "bogus signal",			/* 22 */
  "bogus signal",			/* 23 */
  "bogus signal",			/* 24 */
  "LAN I/O interrupt",			/* 25 SIGAIO */
  "PTY I/O interrupt",			/* 26 SIGPTY */
  "I/O intervention required",		/* 27 SIGIOINT */
  "HFT grant",				/* 28 SIGGRANT */
  "HFT retract",			/* 29 SIGRETRACT */
  "HFT sound done",			/* 30 SIGSOUND */
  "HFT input ready",			/* 31 SIGMSG */
#else /* not IBMRTAIX */
  "bogus signal",			/* 0 */
  "hangup",				/* 1  SIGHUP */
  "interrupt",				/* 2  SIGINT */
  "quit",				/* 3  SIGQUIT */
  "illegal instruction",		/* 4  SIGILL */
  "trace trap",				/* 5  SIGTRAP */
  "IOT instruction",			/* 6  SIGIOT */
  "EMT instruction",			/* 7  SIGEMT */
  "floating point exception",		/* 8  SIGFPE */
  "kill",				/* 9  SIGKILL */
  "bus error",				/* 10 SIGBUS */
  "segmentation violation",		/* 11 SIGSEGV */
  "bad argument to system call",	/* 12 SIGSYS */
  "write on a pipe with no one to read it", /* 13 SIGPIPE */
  "alarm clock",			/* 14 SIGALRM */
  "software termination signum",	/* 15 SIGTERM */
  "user defined signal 1",		/* 16 SIGUSR1 */
  "user defined signal 2",		/* 17 SIGUSR2 */
  "death of a child",			/* 18 SIGCLD */
  "power-fail restart",			/* 19 SIGPWR */
#endif /* not IBMRTAIX */
  0
  };

/*
 *	Warning, this function may not duplicate 4.2 action properly
 *	under error conditions.
 */

#ifndef MAXPATHLEN
/* In 4.1, param.h fails to define this.  */
#define MAXPATHLEN 1024
#endif

#ifndef HAVE_GETWD

char *
getwd (pathname)
     char *pathname;
{
  char *npath, *spath;
  extern char *getcwd ();

  spath = npath = getcwd ((char *) 0, MAXPATHLEN);
  /* On Altos 3068, getcwd can return @hostname/dir, so discard
     up to first slash.  Should be harmless on other systems.  */
  while (*npath && *npath != '/')
    npath++;
  strcpy (pathname, npath);
  free (spath);			/* getcwd uses malloc */
  return pathname;
}

#endif HAVE_GETWD

/*
 *	Emulate rename using unlink/link.  Note that this is
 *	only partially correct.  Also, doesn't enforce restriction
 *	that files be of same type (regular->regular, dir->dir, etc).
 */

rename (from, to)
     char *from;
     char *to;
{
  if (access (from, 0) == 0)
    {
      unlink (to);
      if (link (from, to) == 0)
	if (unlink (from) == 0)
	  return (0);
    }
  return (-1);
}

/* VARARGS */
setpriority ()
{
  return (0);
}

#ifndef HAVE_VFORK

/*
 *	Substitute fork(2) for vfork(2) on USG flavors.
 */

vfork ()
{
  return (fork ());
}

#endif /* not HAVE_VFORK */

#ifdef MISSING_UTIMES

/* HPUX (among others) sets HAVE_TIMEVAL but does not implement utimes.  */

utimes ()
{
}
#endif

#ifdef IRIS_UTIME

/* The IRIS (3.5) has timevals, but uses sys V utime, and doesn't have the
   utimbuf structure defined anywhere but in the man page. */

struct utimbuf
 {
   long actime;
   long modtime;
 };

utimes (name, tvp)
     char *name;
     struct timeval tvp[];
{
  struct utimbuf utb;
  utb.actime  = tvp[0].tv_sec;
  utb.modtime = tvp[1].tv_sec;
  utime (name, &utb);
}
#endif /* IRIS_UTIME */


#ifdef HPUX

/* HPUX curses library references perror, but as far as we know
   it won't be called.  Anyway this definition will do for now.  */

perror ()
{
}

#endif /* HPUX */

#ifndef HAVE_DUP2

/*
 *	Emulate BSD dup2(2).  First close newd if it already exists.
 *	Then, attempt to dup oldd.  If not successful, call dup2 recursively
 *	until we are, then close the unsuccessful ones.
 */

dup2 (oldd, newd)
     int oldd;
     int newd;
{
  register int fd;
  
  sys_close (newd);

#ifdef F_DUPFD
  fd = fcntl (oldd, F_DUPFD, newd);
  if (fd != newd)
    error ("cant dup2(%i,%i) : %s", oldd, newd, sys_errlist[errno]);
#else
  while ((fd = dup (oldd)) != newd)
    {
      dup2 (oldd, newd);
      sys_close (fd);
    }
#endif
}

#endif /* not HAVE_DUP2 */

/*
 *	Gettimeofday.  Simulate as much as possible.  Only accurate
 *	to nearest second.  Emacs doesn't use tzp so ignore it for now.
 *	Only needed when subprocesses are defined.
 */

#ifdef subprocesses
#ifndef HAVE_GETTIMEOFDAY
#ifdef HAVE_TIMEVAL
 
/* ARGSUSED */
gettimeofday (tp, tzp)
     struct timeval *tp;
     struct timezone *tzp;
{
  extern long time ();

  tp->tv_sec = time ((long *)0);    
  tp->tv_usec = 0;
}
 
#endif
#endif
#endif /* subprocess && !HAVE_GETTIMEOFDAY && HAVE_TIMEVAL */
  
/*
 *	This function will go away as soon as all the stubs fixed. (fnf)
 */

croak (badfunc)
     char *badfunc;
{
  printf ("%s not yet implemented\r\n", badfunc);
  reset_sys_modes ();
  exit (1);
}

#endif /* USG */

#ifdef QNX

char *sys_siglist[_SIGMAX + 1] =
{
  "bogus_signal",				/* 0 */
  "hangup",					/* 1 */
  "interrupt",					/* 2 */
  "quit",					/* 3 */
  "illegal instruction",			/* 4 */
  "trace trap",					/* 5 */
  "User abort",					/* 6 */
  "EMT instruction",				/* 7 */
  "floating point exception",			/* 8 */
  "kill",					/* 9 */
  "bus error",					/* 10 */
  "segmentation violation",			/* 11 */
  "bad argument to system call",		/* 12 */
  "write on a pipe with no one to read it",	/* 13 */
  "alarm clock",				/* 14 */
  "software termination signal",		/* 15 */
  "user defined signal 1",			/* 16 */
  "user defined signal 2",			/* 17 */
  "death of a child",				/* 18 */
  "power-fail restart",				/* 19 */
  "window change",				/* 20 */
  "urgent signal",				/* 21 */
  "poll signal",				/* 22 */
  "stop",					/* 23 */
  "terminal stop",				/* 24 */
  "continue",					/* 25 */
  "terminal input",				/* 26 */
  "terminal output",				/* 27 */
  "device signal",				/* 28 */
  "signal 29",					/* 29 */
  "signal 30",					/* 30 */
  "signal 31",					/* 31 */
  "signal 32"					/* 32 */
};

/* Use real free for library functions returning allocated memory */
#undef free

char *
getwd (char *pathname)
{
  char *path;
  path = getcwd((char *) 0, 0);
  strncpy(pathname, path, MAXPATHLEN);
  free(path);			/* Warning: use library free, not sys_free */
  return (pathname);
}

/* Don't bother */
setpriority()
{
  return (0);
}

/* Blecch */
vfork()
{
  return (fork ());
}

/* QNX doesn't have gettimeofday, but it does have gettime. */

int gettimeofday(struct timeval *tv, struct timezone *tz)
{
  struct timespec tp;
  int ret;
  ret=clock_gettime(CLOCK_REALTIME, &tp);
  tv->tv_sec = tp.tv_sec;
  tv->tv_usec = tp.tv_nsec / 1000;
  return(ret);
}

int wait3(status, options, rusage)
int *status;
int options;
char *rusage;			/* not used */
{
  int ret;
  ret = waitpid(-1, status, options);
  return (ret);
}

/*
QNX doesn't have sigsetmask, but they do have sigprocmask, so we do one
in terms of the other.  Why must QNX be different ?
*/

long sigsetmask(long new)
{
  long old=0;
  sigprocmask(SIG_SETMASK, &new, &old);
  return old;
}

croak (badfunc)
char *badfunc;
{
  printf ("%s not yet implemented\r\n", badfunc);
  reset_sys_modes ();
  exit(1);
}

/*
Like mktemp (which doesn't exist) except that we generate the name using
both the node and the process id and we use 10 X's instead of 6.
This guarentees a unique temp name on a network environment.
This will not work if given a node id > 9999.

Example: "/tmp/emacsXXXXXXXXXX" on nid 6 pid 123 becomes "/tmp/emacs000600123A"
*/
char *mktemp_qnx(template)
char *template;
{
  static int unique_id = 'A';
  char tmpbuf[12];
  int i, num;
  char *ptr = template + strlen(template) - 10;
  char *p;
  tmpbuf[0] = '\0';
  ltoa(getnid(), tmpbuf, 10);
  num = strlen(tmpbuf);
  if(num > 4 || num < 1) {
    fprintf(stderr, "Unexpected node id length: %d\n", num);
    abort();		/* Should never happen */
  }
  for(i = 4-num; i>0; i--)
    *ptr++ = '0';		/* Zero prefix */
  p = tmpbuf;
  for(i=num; i>0; i--)
    *ptr++ = *p++;
  itoa(getpid(), tmpbuf, 10);
  num = strlen(tmpbuf);
  if(num > 5 || num < 1) {
    fprintf(stderr, "Unexpected process id length: %d\n", num);
    abort();
  }
  for(i = 5-num; i>0; i--)
    *ptr++ = '0';
  p = tmpbuf;
  for(i=num; i>0; i--)
    *ptr++ = *p++;
  *ptr = unique_id;
  if(unique_id == 'Z')
    unique_id = 'a';
  else if(unique_id == 'z')
    unique_id = 'A';
  else ++unique_id;
  return (template);
}

/*
This provides a combined identifier for processes which may be located on
different nodes.  It allows multiple nodes to use a common lock directory.
Process id's are short ints, so we give 5 decimal digits to it, and the
other digits to the node id.
This is used by filelock.c .

Example: nid 6, pid 1234, gives 601234
*/

int getnidpid()
{
  return(getnid() * 100000 + getpid());
}

/*
Check a given nidpid for existance.
Return 1 if it does, 0 if it doesn't.
If we get an error, return 1 to say that it could exist.
This is used by filelock.c
*/

int checknidpid(nidpid)
int nidpid;
{
  int nid = nidpid / 100000;
  int pid = nidpid % 100000;
  int vid = qnx_vc_attach(nid, pid, 0, 0); /* Try to connect to it */
  /* If we get no error */
  if (vid != -1) {
    qnx_vc_detach(vid);		/* Clean up */
    return(1);			/* Found process */
  } else if (errno == ESRCH)	/* "No such process" */
    return(0);			/* We KNOW it doesn't exist */
  else
    return(1);			/* An error occured, we don't know; be safe */
}

/*
We call this if we run out of stack space.
QNX won't automatically extend the stack.  This is dumb.
Running out of stack space on most operating systems aborts the program.
So it shall under QNX.
Because alloca is called in display code too, we can't just signal an error.
We must die. (Yecch)
Allocating a generous amount of stack space using the cc flag -N <size> should
prevent this error if max-lisp-eval-depth and such are kept reasonable.
*/

void *out_of_stack(void)
{
  write(2, "\r\nOut of stack space -- emacs is dying\r\n", 40);
  abort();				/* This is fatal */
  return(0);				/* Just in case */
}

/*
The conventional abort is virtually useless as a breakpoint.
It is declared as a function that doesn't exit, so it is jumped to and not
called.  It also has no debugging info, and wvideo can't backtrace from
such a function.

Our version prints file and line information, and allows optional return to
the calling function, allowing examination of the condition which caused the
abort.

To make sys_abort return, set abort_debug_return to 1, and step until
sys_abort returns.  Don't go any farther, or else!
*/

static int abort_debug_return;

void sys_abort(char *file, int line)
{
  char buf[20];
  write(2, "\r\nProgram error: ", 17);
  write(2, file, strlen(file));
  write(2, ", line ", 7);
  itoa(line, buf, 10);
  write(2, buf, strlen(buf));
  write(2, "\r\n", 2);
  if(!abort_debug_return)
    abort();
}
#endif

/* Directory routines for systems that don't have them. */

#ifdef SYSV_SYSTEM_DIR

#include <dirent.h>

#ifndef HAVE_CLOSEDIR
int
closedir (dirp)
     register DIR *dirp;              /* stream from opendir() */
{
  sys_close (dirp->dd_fd);
  free ((char *) dirp->dd_buf);       /* directory block defined in <dirent.h> */
  free ((char *) dirp);
}
#endif /* not HAVE_CLOSEDIR */

#endif /* SYSV_SYSTEM_DIR */

#ifdef NONSYSTEM_DIR_LIBRARY

DIR *
opendir (filename)
     char *filename;	/* name of directory */
{
  register DIR *dirp;		/* -> malloc'ed storage */
  register int fd;		/* file descriptor for read */
  struct stat sbuf;		/* result of fstat() */

  fd = sys_open (filename, 0);
  if (fd < 0)
    return 0;

  if (fstat (fd, &sbuf) < 0
      || (sbuf.st_mode & S_IFMT) != S_IFDIR
      || (dirp = (DIR *) malloc (sizeof (DIR))) == 0)
    {
      sys_close (fd);
      return 0;		/* bad luck today */
    }

  dirp->dd_fd = fd;
  dirp->dd_loc = dirp->dd_size = 0;	/* refill needed */

  return dirp;
}

void
closedir (dirp)
     register DIR *dirp;		/* stream from opendir() */
{
  sys_close (dirp->dd_fd);
  free ((char *) dirp);
}


#define DIRSIZ	14
struct olddir
  {
    ino_t od_ino; 		/* inode */
    char od_name[DIRSIZ];	/* filename */
  };

struct direct dir_static;	/* simulated directory contents */

/* ARGUSED */
struct direct *
readdir (dirp)
     register DIR *dirp;	/* stream from opendir() */
{
  register struct olddir *dp;	/* -> directory data */

  for (; ;)
    {
      if (dirp->dd_loc >= dirp->dd_size)
	dirp->dd_loc = dirp->dd_size = 0;

      if (dirp->dd_size == 0 	/* refill buffer */
	  && (dirp->dd_size = sys_read (dirp->dd_fd, dirp->dd_buf, DIRBLKSIZ)) <= 0)
	return 0;

      dp = (struct olddir *) &dirp->dd_buf[dirp->dd_loc];
      dirp->dd_loc += sizeof (struct olddir);

      if (dp->od_ino != 0)	/* not deleted entry */
	{
	  dir_static.d_ino = dp->od_ino;
	  strncpy (dir_static.d_name, dp->od_name, DIRSIZ);
	  dir_static.d_name[DIRSIZ] = '\0';
	  dir_static.d_namlen = strlen (dir_static.d_name);
	  dir_static.d_reclen = sizeof (struct direct)
	    - MAXNAMLEN + 3
	      + dir_static.d_namlen - dir_static.d_namlen % 4;
	  return &dir_static;	/* -> simulated structure */
	}
    }
}

#endif /* NONSYSTEM_DIR_LIBRARY */

#ifdef AIX

/* Get files for keyboard remapping */
#define HFNKEYS 2
#include <sys/hft.h>
#include <sys/devinfo.h>

/* Called from init_sys_modes.  */
hft_init ()
{
  /* On AIX the default hft keyboard mapping uses backspace rather than delete
     as the rubout key's ASCII code.  Here this is changed.  The bug is that
     there's no way to determine the old mapping, so in reset_sys_modes
     we need to assume that the normal map had been present.  Of course, this
     code also doesn't help if on a terminal emulator which doesn't understand
     HFT VTD's. */
  {
    struct hfbuf buf;
    struct hfkeymap keymap;

    buf.hf_bufp = (char *)&keymap;
    buf.hf_buflen = sizeof (keymap);
    keymap.hf_nkeys = 2;
    keymap.hfkey[0].hf_kpos = 15;
    keymap.hfkey[0].hf_kstate = HFMAPCHAR | HFSHFNONE;
    keymap.hfkey[0].hf_page = '<';
    keymap.hfkey[0].hf_char = 127;
    keymap.hfkey[1].hf_kpos = 15;
    keymap.hfkey[1].hf_kstate = HFMAPCHAR | HFSHFSHFT;
    keymap.hfkey[1].hf_page = '<';
    keymap.hfkey[1].hf_char = 127;
    hftctl (0, HFSKBD, &buf);
  }
  /* The HFT system on AIX doesn't optimize for scrolling, so it's really ugly
     at times.  Here we determine if we are on an HFT by trying to get an
     HFT error code.  If this call works, we must be on an HFT. */
  if (ioctl (0, HFQEIO, 0) != -1)
    line_ins_del_ok = char_ins_del_ok = 0;
}

/* Reset the rubout key to backspace. */

hft_reset ()
{
  struct hfbuf buf;
  struct hfkeymap keymap;

  buf.hf_bufp = (char *)&keymap;
  buf.hf_buflen = sizeof(keymap);
  keymap.hf_nkeys = 2;
  keymap.hfkey[0].hf_kpos = 15;
  keymap.hfkey[0].hf_kstate = HFMAPCHAR | HFSHFNONE;
  keymap.hfkey[0].hf_page = '<';
  keymap.hfkey[0].hf_char = 8;
  keymap.hfkey[1].hf_kpos = 15;
  keymap.hfkey[1].hf_kstate = HFMAPCHAR | HFSHFSHFT;
  keymap.hfkey[1].hf_page = '<';
  keymap.hfkey[1].hf_char = 8;
  hftctl (0, HFSKBD, &buf);
}

#endif /* AIX */
